// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

#include "../common.h"

#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <time.h>
#include "../../ms.h"

static int point_compare(void const *p_ptr, void const *q_ptr) {
  uint64_t pr = z_rank(*(struct Point const *)p_ptr);
  uint64_t qr = z_rank(*(struct Point const *)q_ptr);
  if (pr < qr) {
    return -1;
  }
  if (qr < pr) {
    return 1;
  }
  return 0;
}

static int line_compare(void const *l_ptr, void const *m_ptr) {
  struct Line l = *(struct Line const *)l_ptr;
  struct Line m = *(struct Line const *)m_ptr;
  int r = point_compare(&l.p, &m.p);
  return r ? r : point_compare(&l.q, &m.q);
}

static size_t sort_unique(void *base,
                          size_t nmemb,
                          size_t size,
                          int (*compar)(void const *, void const *)) {
  qsort(base, nmemb, size, compar);
  void *a = base;
  for (size_t i = 1; i < nmemb; i++) {
    void const *b = i * size + (char const *)base;
    if (compar(a, b)) {
      a = size + (char *)a;
      memmove(a, b, size);
    }
  }
  return (size_t)((char const *)a - (char const *)base) / size + 1;
}

struct Dart {
  struct Line l;
  int di;
};

static int dart_compare_orientation(void const *d_ptr, void const *e_ptr) {
  struct Dart d = *(struct Dart const *)d_ptr;
  struct Dart e = *(struct Dart const *)e_ptr;
  if (orientation_less(d.l, e.l)) {
    return -1;
  }
  if (orientation_less(e.l, d.l)) {
    return 1;
  }
  return d.di - e.di;
}

static double difftv(struct timeval z, struct timeval a) {
  return difftime(z.tv_sec, a.tv_sec) * 1e6 + ((double)z.tv_usec - a.tv_usec);
}

int main(int argc, char **argv) {
  struct timeval tva;
  gettimeofday(&tva, (struct timezone *)0);
  struct rusage rua;
  getrusage(RUSAGE_SELF, &rua);
  struct Line_list *line_list;
  read_shapefiles(stdin, &line_list);
  int64_t k = 0;
  for (int64_t i = 0; i < line_list->num_lines; i++) {
    struct Line l = line_list->line[i];
    int r = point_compare(&l.p, &l.q);
    if (r < 0) {
      line_list->line[k++] = l;
    } else if (r > 0) {
      line_list->line[k++] = (struct Line){l.q, l.p};
    }
  }
  line_list->num_lines = (int64_t)sort_unique(line_list->line,
                                              (size_t)k,
                                              sizeof *line_list->line,
                                              line_compare);
  assert(line_list->num_lines <= INT64_MAX / 2);
  int64_t num_points = line_list->num_lines * 2;
  struct Point *point = allocate_array(0, (uint64_t)num_points, sizeof *point);
  for (int64_t i = 0; i < line_list->num_lines; i++) {
    struct Line l = line_list->line[i];
    point[i * 2] = l.p;
    point[i * 2 + 1] = l.q;
  }
  num_points = (int64_t)sort_unique(point,
                                    (size_t)num_points,
                                    sizeof *point,
                                    point_compare);
  assert(line_list->num_lines <= INT_MAX / 2 - 1);
  int ne = (int)line_list->num_lines;
  int nd = (1 + ne) * 2;
  struct ms_gr *gr = allocate_array(sizeof *gr, (uint64_t)nd, sizeof *gr->d);
  memset(gr, '\0', sizeof *gr + (unsigned)nd * sizeof *gr->d);
  gr->nv = (int)num_points;
  gr->ne = ne;
  for (int64_t i = 0; i < line_list->num_lines; i++) {
    struct Line l = line_list->line[i];
    struct Point p[2] = {l.p, l.q};
    int64_t length = euclidean_length(l.p, l.q);
    assert(length > 0);
    assert(length <= INT_MAX);
    for (int j = 0; j < 2; j++) {
      int di = (int)((1 + i) * 2 + j);
      gr->d[di].vi = (int)((struct Point const *)bsearch(&p[j],
                                                         point,
                                                         (size_t)num_points,
                                                         sizeof *point,
                                                         point_compare) -
                           point);
      gr->d[di].l = (int)length;
    }
  }
  deallocate(line_list);
  int *vdi = allocate_array(0, (uint64_t)gr->nv, sizeof *vdi);
  memset(vdi, '\0', (unsigned)gr->nv * sizeof *vdi);
  for (int di = 2; di < nd; di++) {
    int vi = gr->d[di].vi;
    gr->d[di].vdi = vdi[vi];
    vdi[vi] = di;
  }
  int max_darts = 0;
  struct Dart *dart = NULL;
  for (int vi = 0; vi < gr->nv; vi++) {
    int num_darts = 0;
    for (int di = vdi[vi]; di; di = gr->d[di].vdi) {
      num_darts++;
    }
    assert(num_darts);
    if (num_darts > max_darts) {
      max_darts = num_darts;
      dart = reallocate_array(dart, 0, (uint64_t)max_darts, sizeof *dart);
    }
    int di = vdi[vi];
    for (int i = 0; i < num_darts; i++) {
      dart[i] = (struct Dart){{point[gr->d[di].vi], point[gr->d[di ^ 1].vi]},
                              di};
      di = gr->d[di].vdi;
    }
    qsort(dart, (size_t)num_darts, sizeof *dart, dart_compare_orientation);
    vdi[vi] = dart->di;
    gr->d[dart[num_darts - 1].di].vdi = dart->di;
    for (int i = 1; i < num_darts; i++) {
      gr->d[dart[i - 1].di].vdi = dart[i].di;
    }
  }
  deallocate(dart);
  for (int di = 2; di < nd; di++) {
    gr->d[gr->d[di].vdi].fi = di;
  }
  for (int vi = 0; vi < gr->nv; vi++) {
    int di = vdi[vi];
    int dj = gr->d[di].vdi;
    if (dj != di && gr->d[dj].vdi == di) {
      gr->d[di].vdi = gr->d[dj ^ 1].vdi == (dj ^ 1) ? di : gr->d[dj ^ 1].vdi;
      gr->d[gr->d[di].vdi].fi = di;
      gr->d[di].vi = gr->d[dj ^ 1].vi;
      vdi[gr->d[di].vi] = di;
      vdi[vi] = 0;
      gr->d[di].fi = gr->d[dj ^ 1].fi == (dj ^ 1) ? di : gr->d[dj ^ 1].fi;
      gr->d[gr->d[di].fi].vdi = di;
      assert(gr->d[dj ^ 1].l <= INT_MAX - gr->d[di].l);
      gr->d[di].l += gr->d[dj ^ 1].l;
      gr->d[dj ^ 1].l = 0;
      assert(gr->d[dj].l <= INT_MAX - gr->d[di ^ 1].l);
      gr->d[di ^ 1].l += gr->d[dj].l;
      gr->d[dj].l = 0;
    }
  }
  int vk = 0;
  for (int vi = 0; vi < gr->nv; vi++) {
    if (vdi[vi]) {
      point[vk] = point[vi];
      vdi[vi] = vk;
      vk++;
    }
  }
  gr->nv = vk;
  if (argc > 1) {
    FILE *stream = fopen(argv[1], "w");
    if (!stream) {
      perror(argv[1]);
      assert(0);
    }
    size_t nmemb = fwrite(point, sizeof *point, (size_t)gr->nv, stream);
    if (nmemb != (unsigned)gr->nv) {
      if (ferror(stream)) {
        perror(argv[1]);
      }
      assert(0);
    }
    if (fclose(stream) == EOF) {
      perror(argv[1]);
      assert(0);
    }
  }
  deallocate(point);
  int dk = 2;
  for (int di = 2; di < nd; di++) {
    if (gr->d[di].l) {
      assert(gr->d[di ^ 1].l);
      gr->d[dk].vdi = gr->d[di].vdi;
      gr->d[dk].vi = vdi[gr->d[di].vi];
      gr->d[dk].l = gr->d[di].l;
      gr->d[di].fi = dk;
      dk++;
    }
  }
  nd = dk;
  gr->ne = (nd >> 1) - 1;
  for (int di = 2; di < nd; di++) {
    gr->d[di].vdi = gr->d[gr->d[di].vdi].fi;
  }
  deallocate(vdi);
  ms_xgr(gr);
  ms_ckgr(gr);
  ms_wrgr(gr, stdout);
  free(gr);
  struct rusage ruz;
  getrusage(RUSAGE_SELF, &ruz);
  struct timeval tvz;
  gettimeofday(&tvz, (struct timezone *)0);
  fprintf(stderr, "{\"user\": %.0f,\n", difftv(ruz.ru_utime, rua.ru_utime));
  fprintf(stderr, " \"system\": %.0f,\n", difftv(ruz.ru_stime, rua.ru_stime));
  fprintf(stderr, " \"elapsed\": %.0f,\n", difftv(tvz, tva));
  fprintf(stderr, " \"major\": %ld,\n", ruz.ru_majflt - rua.ru_majflt);
  fprintf(stderr, " \"maxresident\": %ld}\n", ruz.ru_maxrss);
}
