// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

#include "common.h"

static uint32_t big32_to_host(uint32_t big) {
  unsigned char *s = (unsigned char *)&big;
  return ((uint32_t)(s[3] | ((uint16_t)s[2] << 8)) |
          ((uint32_t)(s[1] | ((uint16_t)s[0] << 8)) << 16));
}

static uint32_t little32_to_host(uint32_t little) {
  unsigned char *s = (unsigned char *)&little;
  return ((uint32_t)(s[0] | ((uint16_t)s[1] << 8)) |
          ((uint32_t)(s[2] | ((uint16_t)s[3] << 8)) << 16));
}

static uint64_t little64_to_host(uint64_t little) {
  unsigned char *s = (unsigned char *)&little;
  return ((uint64_t)((uint32_t)(s[0] | ((uint16_t)s[1] << 8)) |
                     ((uint32_t)(s[2] | ((uint16_t)s[3] << 8)) << 16)) |
          ((uint64_t)((uint32_t)(s[4] | ((uint16_t)s[5] << 8)) |
                      ((uint32_t)(s[6] | ((uint16_t)s[7] << 8)) << 16)) << 32));
}

static struct Point decode_coordinates(uint64_t const coord[2]) {
  return (struct Point){decode_coordinate(little64_to_host(coord[0])),
                        decode_coordinate(little64_to_host(coord[1]))};
}

#define FREAD2(ptr, size)                   \
  do {                                      \
    assert((size) <= file_length);          \
    if (!fread((ptr), (size), 1, stream)) { \
      if (ferror(stream)) {                 \
        perror(NULL);                       \
      }                                     \
      assert(0);                            \
    }                                       \
    file_length -= (size);                  \
  } while (0)

#define FREAD1(array)              \
  FREAD2(&(array), sizeof (array))

void read_shapefiles(FILE *stream, struct Line_list **line_list_ptr) {
  line_list_initialize(line_list_ptr);
  uint32_t *part = NULL;
  uint32_t max_parts = 0;
  uint64_t (*point)[2] = NULL;
  uint32_t max_points = 0;
  while (1) {
    /* HEADER */
    uint32_t header[25];
    size_t bytes_read = fread(header, 1, sizeof header, stream);
    if (bytes_read < sizeof header) {
      if (!bytes_read && feof(stream)) {
        break;
      }
      if (ferror(stream)) {
        perror(NULL);
      }
      assert(0);
    }
    /* File Code */
    assert(big32_to_host(header[0]) == 9994);
    /* File Length */
    uint64_t file_length = (uint64_t)big32_to_host(header[6]) * 2;
    assert(sizeof header <= file_length);
    file_length -= sizeof header;
    /* Version */
    assert(little32_to_host(header[7]) == 1000);
    /* Shape Type */
    enum {
      POLY_LINE = 3
    };
    assert(little32_to_host(header[8]) == POLY_LINE);
    while (file_length > 0) {
      /* RECORD HEADER */
      uint32_t record_header[2];
      FREAD1(record_header);
      /* Content Length */
      uint64_t content_length = (uint64_t)big32_to_host(record_header[1]) * 2;
      /* RECORD CONTENTS */
      uint32_t record_contents[11];
      assert(sizeof record_contents <= content_length);
      FREAD1(record_contents);
      /* Shape Type */
      assert(little32_to_host(record_contents[0]) == POLY_LINE);
      /* NumParts */
      uint32_t num_parts = little32_to_host(record_contents[9]);
      /* NumPoints */
      uint32_t num_points = little32_to_host(record_contents[10]);
      assert(content_length == sizeof record_contents +
             (uint64_t)num_parts * sizeof *part +
             (uint64_t)num_points * sizeof *point);
      assert(num_points >= num_parts);
      /* Parts */
      if (num_parts > max_parts) {
        part = reallocate_array(part, 0, num_parts, sizeof *part);
        max_parts = num_parts;
      }
      FREAD2(part, num_parts * sizeof *part);
      /* Points */
      if (num_points > max_points) {
        point = reallocate_array(point, 0, num_points, sizeof *point);
        max_points = num_points;
      }
      FREAD2(point, num_points * sizeof *point);
      uint32_t j = num_points;
      while (num_parts) {
        uint32_t i = little32_to_host(part[--num_parts]);
        struct Point q = decode_coordinates(point[--j]);
        while (i < j) {
          struct Point p = decode_coordinates(point[--j]);
          line_list_append((struct Line){q, p}, line_list_ptr);
          q = p;
        }
      }
    }
  }
  deallocate(point);
  deallocate(part);
}
