// Copyright 2013--2014 David Eisenstat
// Released under http://opensource.org/licenses/MIT
// mississippi April 2014 version

struct bc_v {
  struct bc_v *p;
  struct bc_v *l;
  struct bc_v *r;
  int a12;
  int a23;
  unsigned long long a13;
};

static void bc_attach(struct bc_v *restrict u, struct bc_v const *restrict v) {
  u->a12 -= v->a12;
  u->a23 -= v->a23;
  u->a13 -= v->a13 + (unsigned long long)u->a12 * (unsigned long long)v->a23;
}

static void bc_detach(struct bc_v *restrict u, struct bc_v const *restrict v) {
  u->a13 += v->a13 + (unsigned long long)u->a12 * (unsigned long long)v->a23;
  u->a23 += v->a23;
  u->a12 += v->a12;
}

static void bc_attach_bastard(struct bc_v *restrict u,
                              struct bc_v const *restrict v) {
  u->a23 -= v->a23;
  u->a13 -= (unsigned long long)u->a12 * (unsigned long long)v->a23;
}

static void bc_detach_bastard(struct bc_v *restrict u,
                              struct bc_v const *restrict v) {
  u->a13 += (unsigned long long)u->a12 * (unsigned long long)v->a23;
  u->a23 += v->a23;
}

#define BC_ROTATE(L, R)                                        \
  static void bc_##L##rotate(struct bc_v *u, struct bc_v *v) { \
    v->R = u->L;                                               \
    u->L = v;                                                  \
    v->p = u;                                                  \
    bc_attach(v, u);                                           \
    if (v->R) {                                                \
      v->R->p = v;                                             \
      bc_attach(v->R, v);                                      \
    }                                                          \
  }

BC_ROTATE(l, r)
BC_ROTATE(r, l)

#define BC_2ROTATE(L, R)    \
  w = v->p;                 \
  if (w) {                  \
    struct bc_v *x = w->p;  \
    if (w->R == v) {        \
      bc_detach(v, w);      \
      bc_detach(u, v);      \
      bc_##L##rotate(v, w); \
      bc_##L##rotate(u, v); \
      v = x;                \
      continue;             \
    } else if (w->L == v) { \
      bc_detach(v, w);      \
      bc_detach(u, v);      \
      bc_##L##rotate(u, v); \
      bc_##R##rotate(u, w); \
      v = x;                \
      continue;             \
    }                       \
  }                         \
  u->p = w;                 \
  bc_detach(u, v);          \
  bc_##L##rotate(u, v);     \
  break

static void bc_splay(struct bc_v *u) {
  struct bc_v *v = u->p;
  struct bc_v *w = u;
  while (1) {
    if (v) {
      if (v->r == w) {
        BC_2ROTATE(l, r);
      } else if (v->l == w) {
        BC_2ROTATE(r, l);
      }
    }
    u->p = v;
    break;
  }
}

static void bc_expose(struct bc_v *u) {
  bc_splay(u);
  struct bc_v *v = u->p;
  u->p = 0;
  while (v) {
    bc_splay(v);
    if (v->l) {
      bc_detach(v->l, v);
      bc_attach_bastard(v->l, v);
    }
    struct bc_v *w = v->p;
    bc_detach_bastard(u, v);
    bc_rrotate(u, v);
    v = w;
  }
}

static void bc_initialize(struct bc_v *u) {
  u->p = 0;
  u->l = 0;
  u->r = 0;
  u->a12 = 1;
}

static void bc_new_root(struct bc_v *u) {
  bc_expose(u);
  u->a13 -= (unsigned long long)u->a12 - 1;
  u->a23++;
  if (u->l) {
    u->l->a13 += 2 * (unsigned long long)u->a12 - 1;
  }
  if (u->r) {
    u->r->a13 += 2 * (unsigned long long)u->a12 - 1;
  }
}

static void bc_link(struct bc_v *u, struct bc_v *v) {
  bc_expose(u);
  bc_expose(v);
  if (v->l) {
    bc_detach(v->l, v);
    bc_attach_bastard(v->l, v);
  }
  v->a12 += u->a12;
  v->l = u;
  u->p = v;
  bc_attach(u, v);
}

static void bc_cut(struct bc_v *u) {
  bc_expose(u);
  bc_detach(u->r, u);
  u->r->a12 -= u->a12;
  u->r->p = 0;
  u->r = 0;
}

static unsigned long long bc_centrality(struct bc_v *u) {
  bc_expose(u);
  return u->a13;
}

struct bc_args {
  struct ms_gr const *gr;
  struct bc_v *v;
  unsigned long long np;
};

static void bc_on_pivot(void *o, int cdi, int ldi) {
  struct bc_args *args = (struct bc_args *)o;
  if (cdi) {
    bc_cut(&args->v[args->gr->d[cdi].vi]);
  }
  if (ldi) {
    bc_link(&args->v[args->gr->d[ldi].vi], &args->v[args->gr->d[ldi ^ 1].vi]);
  }
  args->np++;
}
